/* (c) Copyright 2001 Josh Levenberg */

#include <assert.h>
#include <math.h>
#include "collide.h"
#include "world.h"
#include "glswmath.h"
#include "glutmain.h"

/* Tests shot for intersection vs. a wall.
   If this closer than previous intersections, use this intersection instead.
 */
static bool TestForCloserIntersection(float raydir[3], float normal[3], float offset)
  {
    float intersect[3];
    bool hit=RayPlaneIntersection(gPlayerPos, raydir, normal, offset, intersect);
    if (hit && intersect[1] < gShotDest[1])
      {
        gShotDest[0]=intersect[0];
        gShotDest[1]=intersect[1];
        gShotDest[2]=intersect[2];
      }
    return hit;
  }

/* Efficient shot vs. gun port collision detection.
   Efficient because it first culls all but a rectangle of gun ports
   before performing ray vs. sphere intersection tests.
   Here "c" refers to x or z, depending on if the gun port is on
   the floor or a wall.
 */
static void TestForGunPortHit(int maxy, float raydir[3], float normal[3], 
                              float offset, int minc, int maxc, int ctowhich,
                              float gunporthitlocation[3], int gunporthit[2])
  {
    int miny, y, c;
    float intersect[3];
    
    if (RayPlaneIntersection(gPlayerPos, raydir, normal,
                              offset+GUN_RADIUS, intersect))
      miny = (int) (intersect[1]/TRENCH_DEPTH);
    else
      miny = 0;

    for (y=miny; y<=maxy; ++y)
      for (c=minc; c<=maxc; ++c)
        if (gGunPorts[y] & (1<<(c+ctowhich)))
          {
            float gunpos[3] = 
              { gGunPortXZ[c+ctowhich][0], 
                TRENCH_DEPTH*y+GUN_PORT_Y, 
                gGunPortXZ[c+ctowhich][1] 
              };
            float intersect[3];
            if (LineSphereIntersect(gPlayerPos, raydir, gunpos,
                        GUN_RADIUS*GUN_RADIUS, intersect))
              {
                if (intersect[1] < gShotDest[1])
                  { /* Hit! */
                    int i;
                    for (i=0; i<3; ++i)
                      {
                        gShotDest[i]=intersect[i];
                        gunporthitlocation[i]=gunpos[i];
                      }
                    gunporthit[0]=y;
                    gunporthit[1]=c+ctowhich;
                  }
              }
          }
  }

void ComputeShotDestination()
  {
    /* Goal: compute gShotDest from gPlayerPos and gShootAim */
    float raydir[3];
    static float normals[5][3] =
      { { 0, -1, 0 }, /* Far plane */
        { 0, 0, -1 }, /* "Top" plane */
        { 0, 0, 1 },  /* Bottom plane -- floor */
        { 1, 0, 0 },  /* Left plane */
        { -1, 0, 0 } /* Right plane */
      };
    float offsets[5] =
      { -gFar, -TRENCH_TOP, TRENCH_BOTTOM, TRENCH_LEFT, -TRENCH_RIGHT };
    bool hit;
    int i;
    int gunporthit[2] = {-1,-1};
    int missilehit = -1;
    float hitlocation[3];

    raydir[1] = (float) (.5/tan(DegreesToRadians(gFov/2)));
    /* multiplying by .5 is the same as dividing by 2 
      (there is no roundoff problem for powers of 2) */
    raydir[0] = (float) ((gShootAim[0]-.5)*gWidth/gHeight);
    raydir[2] = (float) (gShootAim[1]-.5);

    NormalizeVector(raydir);

    /* check where ray hits far plane */
    hit=RayPlaneIntersection(gPlayerPos, raydir, normals[0], offsets[0], gShotDest);
    assert(hit);

    /* --- see if hits anything closer --- */
    /* walls, floor, "ceiling" */
    for (i=1; i<=4; ++i)
      TestForCloserIntersection(raydir, normals[i], offsets[i]);

    /* explosions */
    for (i=0; i<EXPLOSIONS_MAX; ++i)
      if (gExplosions[i].Countdown>0)
        {
          float intersect[3];
          if (LineSphereIntersect(gPlayerPos, raydir, gExplosions[i].Pos,
                      EXPLOSIONS_RADIUS*EXPLOSIONS_RADIUS, intersect))
            {
              if (intersect[1] < gShotDest[1])
                { /* Hit! */
                  int j;
                  for (j=0; j<3; ++j)
                    gShotDest[j]=intersect[j];
                }
            }
        }

    { /* check for gun port hits */
      int maxy = (int) (gShotDest[1]/TRENCH_DEPTH);

      { /* left and right walls */

        /* check which zs could be hit */
        int minz, maxz;
        if (raydir[2] < 0)
          { /* ray pointing down */
            minz=0;
            for (maxz=GUN_PORT_PER_WALL-1; maxz>=0; --maxz)
              if (gPlayerPos[2]>gGunPortXZ[maxz][1]-GUN_RADIUS)
                break;
          }
        else
          { /* ray pointing up */
            maxz=2;
            for (minz=0; minz<GUN_PORT_PER_WALL; ++minz)
              if (gPlayerPos[2]<gGunPortXZ[minz][1]+GUN_RADIUS)
                break;
          }

        if (raydir[0] < 0 || gPlayerPos[0] < TRENCH_LEFT+GUN_RADIUS)
          { /* check left wall */
            TestForGunPortHit(maxy, raydir, normals[3], offsets[3], 
                              minz, maxz, 0, hitlocation,
                              gunporthit);
          }
        if (raydir[0] > 0 || gPlayerPos[0] > TRENCH_RIGHT-GUN_RADIUS)
          { /* check right wall */
            TestForGunPortHit(maxy, raydir, normals[4], offsets[4], 
                              minz, maxz, 6, hitlocation,
                              gunporthit);
          }
      }
      if (raydir[2] < 0 || gPlayerPos[2] < TRENCH_BOTTOM+GUN_RADIUS)
        { /* check floor */
          int minx, maxx;

          if (raydir[0] < 0)
            { /* ray pointing left */
              minx=0;
              for (maxx=2; maxx>=0; --maxx)
                if (gPlayerPos[0]>gGunPortXZ[maxx+3][0]-GUN_RADIUS)
                  break;
            }
          else
            { /* ray pointing right */
              maxx=2;
              for (minx=0; minx<=2; ++minx)
                if (gPlayerPos[0]<gGunPortXZ[minx+3][0]+GUN_RADIUS)
                  break;
            }

          TestForGunPortHit(maxy, raydir, normals[2], offsets[2], 
                            minx, maxx, 3, hitlocation,
                            gunporthit);
        }
    }

    /* missiles */
    for (i=0; i<MISSILES_MAX; ++i)
      if (gMissiles[i].Enabled)
        {
          float intersect[3];
          if (LineSphereIntersect(gPlayerPos, raydir, gMissiles[i].Pos,
                      MISSILE_RADIUS*MISSILE_RADIUS, intersect))
            {
              if (intersect[1] < gShotDest[1])
                { /* Hit! */
                  int j;
                  for (j=0; j<3; ++j)
                    gShotDest[j]=intersect[j];
                  missilehit = i;
                  gunporthit[0] = -1; /* supercede any gun port hit */
                }
            }
        }


    if (gunporthit[0]>-1)
      {
        /* explosion will be hit on future cycles
           need: explosion display time longer than firing time
                 but shorter than firing time + cooldown time
           need: explosion radius larger than gun port radius
                 and missile radius
           need: insure only one explosion at a time (since have
                 only a single light source to spare)
         */
        if (gunporthit[1]<GUN_PORT_PER_WALL)
          CreateExplosion(hitlocation, EXPLODE_RIGHT);
        else if (gunporthit[1]<2*GUN_PORT_PER_WALL)
          CreateExplosion(hitlocation, EXPLODE_UP);
        else
          CreateExplosion(hitlocation, EXPLODE_LEFT);
        gGunPorts[gunporthit[0]] &= GUN_PORT_MASK - (1<<gunporthit[1]);
        gScore+=GUN_PORT_SCORE;
      }
    else if (missilehit>-1)
      {
        CreateExplosion(gMissiles[missilehit].Pos, EXPLODE_ALL);
        gMissiles[missilehit].Enabled=false;
        gScore+=MISSILE_SCORE;
      }
  }

bool DetectCollision(float oldy)
  {
    bool ret=false;
    int missile;
    for (missile=0; missile<MISSILES_MAX; ++missile)
      /* check for trivial regection */
      if (gMissiles[missile].Enabled &&
          gMissiles[missile].Pos[1]-MISSILE_RADIUS < 
                                    gPlayerPos[1]+PLAYER_COLLISION_Y &&
          gMissiles[missile].Pos[1]+MISSILE_RADIUS > 
                                    oldy+PLAYER_COLLISION_Y)
        { /* potential intersection */
          int i;
          float test[3];
          float dist[3];

          /* compute test, the closest feature to the center of the missile */
          if (gMissiles[missile].Pos[0]>gPlayerPos[0]+PLAYER_WIDTH*.5f)
            test[0]=gPlayerPos[0]+PLAYER_WIDTH*.5f;
          else if (gMissiles[missile].Pos[0]<gPlayerPos[0]-PLAYER_WIDTH*.5f)
            test[0]=gPlayerPos[0]+PLAYER_WIDTH*.5f;
          else
            test[0]=gMissiles[missile].Pos[0];

          if (gMissiles[missile].Pos[1]>gPlayerPos[1]+PLAYER_COLLISION_Y)
            test[1]=gPlayerPos[1]+PLAYER_COLLISION_Y;
          else if (gMissiles[missile].Pos[1]<oldy+PLAYER_COLLISION_Y)
            test[1]=oldy+PLAYER_COLLISION_Y;
          else
            test[1]=gMissiles[missile].Pos[1];

          if (gMissiles[missile].Pos[2]>gPlayerPos[2]+PLAYER_HEIGHT*.5f)
            test[2]=gPlayerPos[2]+PLAYER_HEIGHT*.5f;
          else if (gMissiles[missile].Pos[2]<gPlayerPos[2]-PLAYER_HEIGHT*.5f)
            test[2]=gPlayerPos[2]+PLAYER_HEIGHT*.5f;
          else
            test[2]=gMissiles[missile].Pos[2];

          /* compute dist, the vector between test and 
              the center of the missile */
          for (i=0; i<3; ++i)
            dist[i]=gMissiles[missile].Pos[i]-test[i];

          /* is dist longer than the radius of the missile? */
          if (DotProduct(dist, dist) < MISSILE_RADIUS*MISSILE_RADIUS)
            { /* collision! */
              CreateExplosion(gMissiles[missile].Pos, EXPLODE_ALL);
              gMissiles[missile].Enabled=false;
              ret=true;
            }
        }
    if (ret)
      return true;
    if (gObstacles[0] && 
        oldy+PLAYER_COLLISION_Y <= OBSTACLE_Y && 
        gPlayerPos[1]+PLAYER_COLLISION_Y >= OBSTACLE_Y)
      {
        bool occupied[OBSTACLES_COLUMNS][OBSTACLES_ROWS];
        int row, col;
        /* determine which cells are occupied by obstacle */
        switch (gObstacles[0] & OBSTACLES_TYPE_MASK)
          {
            case OBSTACLES_HORIZ:
              for (row=0; row<OBSTACLES_ROWS; ++row)
                if (gObstacles[0] & (1<<row))
                  for (col=0; col<OBSTACLES_COLUMNS; ++col)
                    occupied[col][row]=true;
                else
                  for (col=0; col<OBSTACLES_COLUMNS; ++col)
                    occupied[col][row]=false;
              break;

            case OBSTACLES_VERT:
              for (col=0; col<OBSTACLES_COLUMNS; ++col)
                if (gObstacles[0] & (1<<col))
                  for (row=0; row<OBSTACLES_ROWS; ++row)
                    occupied[col][row]=true;
                else
                  for (row=0; row<OBSTACLES_ROWS; ++row)
                    occupied[col][row]=false;
              break;

            case OBSTACLES_WALL:
              for (row=0; row<OBSTACLES_ROWS; ++row)
                for (col=0; col<OBSTACLES_COLUMNS; ++col)
                  if (gObstacles[0] & (1<<(col+row*OBSTACLES_COLUMNS)))
                    occupied[col][row]=true;
                  else
                    occupied[col][row]=false;
              break;

            default:
              assert(0); /* should never get here */
          }
        /* test if the cells the player occupies have obstacles */  
          {
            float player_left    = gPlayerPos[0] - PLAYER_WIDTH/2 - TRENCH_LEFT;
            float player_right   = player_left+PLAYER_WIDTH;
            float player_bottom  = gPlayerPos[2] - PLAYER_HEIGHT/2 - TRENCH_BOTTOM;
            float player_top     = player_bottom+PLAYER_HEIGHT;
            int cell_left   = (int) (player_left  /OBSTACLE_WIDE);
            int cell_right  = (int) (player_right /OBSTACLE_WIDE);
            int cell_bottom = (int) (player_bottom/OBSTACLE_HIGH);
            int cell_top    = (int) (player_top   /OBSTACLE_HIGH);
            for (row=cell_bottom; row<=cell_top; ++row)
              for (col=cell_left; col<=cell_right; ++col)
                if (occupied[col][row])
                  return true;
          }
      }

    return false;
  }

