/* (c) Copyright 2001 Josh Levenberg */

#include <stdlib.h>
#include <math.h>
#include "glswmath.h"

/* for radian conversion */
#define PI  (3.1415926535897932384626)
/* for ray-plane intersection */
#define EPSILON  (0.0000000001)

double DegreesToRadians(float foo)
  {
    return foo * PI / 180;
  }

float DotProduct(float v1[3], float v2[3])
  {
    return v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2];
  }

void NormalizeVector(float v[3])
  {
    float oneoverlength = (float) (1.f/sqrt(DotProduct(v, v)));
    int i;
    for (i=0; i<3; ++i)
      v[i] *= oneoverlength;
  }

/*
  |   i     j     k   |
  | v1[0] v1[1] v1[2] |
  | v2[0] v2[1] v2[2] |
*/
void CrossProduct(float v1[3], float v2[3], float out_v[3])
  {
    out_v[0] = - v1[1]*v2[2] + v1[2]*v2[1];
    out_v[1] =   v1[0]*v2[2] - v1[2]*v2[0];
    out_v[2] = - v1[0]*v2[1] + v1[1]*v2[0];
  }

/*
  Plane defined by Dot(P, normal) = offset
  Ray starts at pos and extends in the direction indicated by dir

   Origin            |               dir
      +              |                \  Ray
       <-- offset -->|--> Normal       \
                     |                  + Pos
                    Plane
*/
bool RayPlaneIntersection(float pos[3], float dir[3], float normal[3], float offset, float result[3])
  {
    float scalef;
    float dotprod = DotProduct(dir,normal);
    /* check to see if ray is going away from plane */
    if (dotprod >= -EPSILON)
      return false;
    scalef = -(DotProduct(pos,normal)-offset)/dotprod;
    result[0] = pos[0]+scalef*dir[0];
    result[1] = pos[1]+scalef*dir[1];
    result[2] = pos[2]+scalef*dir[2];
    return true;
  }

/* returns a number from 0..n-1 */
int Random(int n)
  {
    return (rand()*n)/(RAND_MAX+1);
  }

/* returns a floating point number from 0..1 */
float FRand()
  {
    return ((float) rand())/RAND_MAX;
  }

/* Determines the squared distance from a point to a line.
   Used by LineSphereIntersect
   Assumes linedir is normalized 
 */
static float DistancePointLineSquared(float linepos[3], float linedir[3], 
                        float point[3], float out_closestpt[3])
  {
    float vect[3], dist[3];
    float projection;
    int i;
    for (i=0; i<3; ++i)
      vect[i]=point[i]-linepos[i];
    projection=DotProduct(vect, linedir);
    /* closest point on line is now linepos+projection*linedir */
    for (i=0; i<3; ++i)
      out_closestpt[i]=linepos[i]+projection*linedir[i];
    /* so compute the distance between point and closest point */
    for (i=0; i<3; ++i)
      dist[i]=point[i]-out_closestpt[i];
    return DotProduct(dist, dist);
  }

bool LineSphereIntersect(float linepos[3], float linedir[3], 
                        float center[3], float radiussquared, 
                        float out_closestpt[3])
  {
    float distsq=DistancePointLineSquared(linepos, linedir, center, 
                                          out_closestpt);
    if (distsq<radiussquared)
      /* intersected, now back up to find first intersection */
      {
        float backup = (float) sqrt(radiussquared - distsq);
        int i;
        for (i=0; i<3; ++i)
          out_closestpt[i] -= backup*linedir[i];
        return true;
      }
    return false;
  }
