/* (c) Copyright 2001 Josh Levenberg */

/* Has code for rendering the models of the game.  
   Used by Initialize() to compile display lists
   (so performance here is not as important as it may seem).

   Note that all lighting is precomputed ("baked in")
   at this stage.
*/

#include <GL/glut.h>
#include <math.h>
#include "model.h"
#include "world.h"
#include "glswmath.h"

unsigned gModelDisplayList;

static void DrawTrenchModel(void);
static void DrawGunModel(void);
static void DrawGunModelLeft(void);
static void DrawGunModelFloor(void);
static void DrawGunModelRight(void);
static void DrawMissileModel(void);

voidfunc gDrawModelFuncArray[NumModels] =
  {
    DrawTrenchModel,
    DrawGunModelLeft,
    DrawGunModelFloor,
    DrawGunModelRight,
    DrawMissileModel
  };

/* Coordinate system:
   +x right
   +y forward
   +z up
 */

/* Lighting computation.
   Slight cheating: DotProduct should be clipped at 0
   not -.1, but this gives a slightly larger range of shades
 */

static float sDefaultLightVector[3] = 
  { -1.f*0.15275f, -2.f*0.15275f, 4.f*0.15275f };

static float ComputeLighting(float verts[3][3])
  {
    float vects[2][3];
    float normal[3];
    int i;
    float ret;
    for (i=0; i<3; ++i)
      {
        vects[0][i]=verts[1][i]-verts[0][i];
        vects[1][i]=verts[2][i]-verts[0][i];
      }
    CrossProduct(vects[1], vects[0], normal);
    NormalizeVector(normal);
    ret = DotProduct(normal, sDefaultLightVector);
    if (ret<-.1f)
      ret=-.1f;
    return ret+.3f;
  }

/* Renders a strip with lighting added in */

static void RenderStrip(float clr[3], float verts[][3], unsigned short inds[], int numind)
  {
    int firstv;
    for (firstv=0; firstv < numind-2; ++firstv)
      {
        int corn_ind[3];
        float corn_coords[3][3];
        float lighting;
        float newclr[3];
        int i,j;

        if (firstv & 1)
          {
            corn_ind[0]=inds[firstv+1];
            corn_ind[1]=inds[firstv];
            corn_ind[2]=inds[firstv+2];
          }
        else
          {
            corn_ind[0]=inds[firstv];
            corn_ind[1]=inds[firstv+1];
            corn_ind[2]=inds[firstv+2];
          }

        for (i=0; i<3; ++i)
          for (j=0; j<3; ++j)
            corn_coords[i][j] = verts[ corn_ind[i] ][j];
        lighting=ComputeLighting(corn_coords);
        for (i=0; i<3; ++i)
          newclr[i]=clr[i]*lighting;
        glColor3fv(newclr);
        for (i=0; i<3; ++i)
          glArrayElement(corn_ind[i]);
      }
  }

/* Renders a fan with lighting added in */

static void RenderFan(float clr[3], float verts[][3], unsigned short inds[], int numind)
  {
    int firstv;
    for (firstv=1; firstv < numind-1; ++firstv)
      {
        int corn_ind[3];
        float corn_coords[3][3];
        float lighting;
        float newclr[3];
        int i,j;

        corn_ind[0]=inds[0];
        corn_ind[1]=inds[firstv];
        corn_ind[2]=inds[firstv+1];

        for (i=0; i<3; ++i)
          for (j=0; j<3; ++j)
            corn_coords[i][j] = verts[ corn_ind[i] ][j];
        lighting=ComputeLighting(corn_coords);
        for (i=0; i<3; ++i)
          newclr[i]=clr[i]*lighting;
        glColor3fv(newclr);
        for (i=0; i<3; ++i)
          glArrayElement(corn_ind[i]);
      }
  }

/* Used for one segment of the trench.  
   Could easily switch this to be a single model for the entire trench.
   Could probably also remove the backfacing half of the ridge.
*/

static void DrawTrenchModel()
  {
#define RIDGE_FORE     (0.90f)
#define RIDGE_MID      (0.95f)
#define RIDGE_AFT      (1.0f)
#define RIDGE_DENT     (0.05f)
#define TRENCH_AFT     RIDGE_FORE

    float verts[][3] = 
      { { TRENCH_RIGHT, TRENCH_FORE, TRENCH_TOP },      /* 0 */
        { TRENCH_RIGHT, TRENCH_FORE, TRENCH_BOTTOM  },  /* 1 */ 
        { TRENCH_LEFT,  TRENCH_FORE, TRENCH_BOTTOM  },  /* 2 */ 
        { TRENCH_LEFT,  TRENCH_FORE, TRENCH_TOP },      /* 3 */ 
        { TRENCH_RIGHT, TRENCH_AFT,  TRENCH_TOP },      /* 4 */
        { TRENCH_RIGHT, TRENCH_AFT,  TRENCH_BOTTOM  },  /* 5 */ 
        { TRENCH_LEFT,  TRENCH_AFT,  TRENCH_BOTTOM  },  /* 6 */ 
        { TRENCH_LEFT,  TRENCH_AFT,  TRENCH_TOP },      /* 7 */ 
        /* (dent ridge outwards) */
        { TRENCH_RIGHT+RIDGE_DENT, RIDGE_MID, TRENCH_TOP },               /* 8 */
        { TRENCH_RIGHT+RIDGE_DENT, RIDGE_MID, TRENCH_BOTTOM-RIDGE_DENT }, /* 9 */
        { TRENCH_LEFT -RIDGE_DENT, RIDGE_MID, TRENCH_BOTTOM-RIDGE_DENT }, /* 10 */
        { TRENCH_LEFT -RIDGE_DENT, RIDGE_MID, TRENCH_TOP },               /* 11 */
        { TRENCH_RIGHT, RIDGE_AFT,  TRENCH_TOP },       /* 12 */
        { TRENCH_RIGHT, RIDGE_AFT,  TRENCH_BOTTOM  },   /* 13 */ 
        { TRENCH_LEFT,  RIDGE_AFT,  TRENCH_BOTTOM  },   /* 14 */ 
        { TRENCH_LEFT,  RIDGE_AFT,  TRENCH_TOP }        /* 15 */ 
      };

    /* 7_5_3_1  Well, actually: 7     1
       |/|/|/|                  |\   /|
       6-4-2-0                  6-5-3 0
                                 \|/|/
                                  4-2
     */
    unsigned short strip1[] = { 0, 4, 1, 5, 2, 6, 3, 7 };
    unsigned short strip2[] = { 4, 8, 5, 9, 6, 10, 7, 11 };
    unsigned short strip3[] = { 8, 12, 9, 13, 10, 14, 11, 15 };
    float clrs[2][3] = 
      { { .4f, .4f, .45f },   /* slightly bluish gray */
        { .45, .45, .4f  } }; /* slightly yellowish gray */

    /* FUTURE: add normals for lighting */
    glVertexPointer(3, GL_FLOAT, 0, verts);
    
    glBegin(GL_TRIANGLES);
    /* Main trench walls */

    RenderStrip(clrs[0], verts, strip1, sizeof(strip1)/sizeof(strip1[0]));

    /* Ridge dividing segments */
    /* Remove backfacing faces:
    RenderStrip(clrs[1], verts, strip2, sizeof(strip2)/sizeof(strip2[0]));
    */
    RenderStrip(clrs[1], verts, strip3, sizeof(strip2)/sizeof(strip3[0]));

    glEnd();
  }

/* Used for gun port.  Note that these occur with three different
   orientations -- to get the lighting right, need to make three
   display lists.
*/

static void DrawGunModelRotated(int howrotated)
  {
    int i;
    float verts[34][3];
    unsigned short strip1[12] =
      { 5, 0, 6, 1, 7, 2, 8, 3, 9, 4, 5, 0 };
    unsigned short fan1[5] =
      { 5, 6, 7, 8, 9 };
    unsigned short strips[3][14] =
      { { 16, 10, 17, 11, 18, 12, 19, 13, 20, 14, 21, 15, 16, 10 },
        { 22, 16, 23, 17, 24, 18, 25, 19, 26, 20, 27, 21, 22, 16 },
        { 28, 22, 29, 23, 30, 24, 31, 25, 32, 26, 33, 27, 28, 22 } };

    float clrs[4][3] =
      { { 1.0f, 0.4f, 0.0f },  /* orangish red */
        { 1.0f, 0.2f, 0.0f },  /* more red */
        { 1.0f, 0.5f, 0.3f },  /* brighter */
        { 0.7f, 0.1f, 0.0f }   /* darker */
      };

    for (i=0; i<5; ++i)
      {
        float radians = DegreesToRadians(i*(360/5));        
        verts[i][0] =  GUN_RADIUS*sin(radians);
        verts[i][1] = -GUN_RADIUS*cos(radians);
        verts[i][2] =  0.0;
        verts[i+5][0]=verts[i][0];
        verts[i+5][1]=verts[i][1];
        verts[i+5][2]=GUN_RADIUS/3;
      }

    for (i=0; i<6; ++i)
      {
        float radians = DegreesToRadians(i*(360/6));
        float cosangle= cos(radians);
        float sinangle= sin(radians);

        /* Outside base */
        verts[i+10][0]=(GUN_RADIUS/3)    *cosangle;
        verts[i+10][1]=(GUN_RADIUS/5)    *sinangle+(GUN_RADIUS*.6);
        verts[i+10][2]=(GUN_RADIUS/6)    *sinangle+(GUN_RADIUS/6);
        /* Outside top */
        verts[i+16][0]=(GUN_RADIUS/3)    *cosangle;
        verts[i+16][1]=(GUN_RADIUS/5)    *sinangle-(GUN_RADIUS*2)/3;
        verts[i+16][2]=(GUN_RADIUS/10)   *sinangle+(GUN_RADIUS*2)/3;
        /* Inside top */
        verts[i+22][0]=(GUN_RADIUS/3) *.7*cosangle;
        verts[i+22][1]=(GUN_RADIUS/5) *.7*sinangle-(GUN_RADIUS*2)/3;
        verts[i+22][2]=(GUN_RADIUS/10)*.7*sinangle+(GUN_RADIUS*2)/3;
        /* Inside base */
        verts[i+28][0]=(GUN_RADIUS/3) *.7*cosangle;
        verts[i+28][1]=(GUN_RADIUS/5) *.7*sinangle+(GUN_RADIUS*.6);
        verts[i+28][2]=(GUN_RADIUS/6) *.7*sinangle+(GUN_RADIUS/6);
      }

    /* rotate vertices */
    if (howrotated)
      for (i=0; i<sizeof(verts)/sizeof(verts[0]); ++i)
        {
          float temp = verts[i][0];
          verts[i][0] = -howrotated*verts[i][2];
          verts[i][2] = howrotated*temp;
        }
      
    glVertexPointer(3, GL_FLOAT, 0, verts);
    
    glBegin(GL_TRIANGLES);
    /* Base */
    RenderStrip(clrs[0], verts, strip1, sizeof(strip1)/sizeof(strip1[0]));
    RenderFan(clrs[0], verts, fan1, sizeof(fan1)/sizeof(fan1[0]));

    /* Barrel */
    RenderStrip(clrs[1], verts, strips[0], sizeof(strips[0])/sizeof(strips[0][0]));
    RenderStrip(clrs[2], verts, strips[1], sizeof(strips[1])/sizeof(strips[1][0]));
    RenderStrip(clrs[3], verts, strips[2], sizeof(strips[2])/sizeof(strips[2][0]));
    glEnd();
  }

static void DrawGunModelLeft()
  {
    DrawGunModelRotated(-1);
  }

static void DrawGunModelFloor()
  {
    DrawGunModelRotated(0);
  }

static void DrawGunModelRight()
  {
    DrawGunModelRotated(1);
  }

/* Missiles are assumed to be emmisive and therefore don't
   use the lighting support functions above.
   Missiles are modelled as a stellated icosahedron 
 */
static void DrawMissileModel()
  { 
    int i, j, k;
    float verts[12+20][3];
    float golden_ratio = (1.f+sqrt(5.f))/2.f;
    /* I figured out how the faces worked on scratch paper */
    int twentyfaces[20][3]=
      { { 0,1,6 },  { 1,0,4 },  { 2,3,5 },   { 3,2,7 }, 
        { 4,5,10 }, { 5,4,8 },  { 6,7,9 },   { 7,6,11 },
        { 8,9,2 },  { 9,8,0 },  { 10,3,11 }, { 11,1,10 },
        { 3,7,11 }, { 1,11,6 }, { 2,9,7 },   { 0,6,9 },
        { 3,10,5 }, { 1,4,10 }, { 2,5,8 },   { 0,8,4 }
      };

#define VERTEX_SCALE   (MISSILE_RADIUS/1.9f/1.2f)
    for (i=0; i<3; ++i) /* determines the order of 0, 1, golden_ratio */
      for (j=0; j<2 ; ++j) /* determines the sign of 1 */
        for (k=0; k<2; ++k) /* determines the sign of golden_ratio */
          {
            int vertnum=i*4+2*j+k;
            verts[vertnum][(i+0)%3]=0;     
                  /* (i+0)%3==i but speed not important here */
            verts[vertnum][(i+1)%3]=(2*k-1)*VERTEX_SCALE;
                                        /* 2*k-1 == +/- 1 */
            verts[vertnum][(i+2)%3]=(2*j-1)*golden_ratio*VERTEX_SCALE;
                                        /* 2*j-1 == +/- 1 */
          }
#undef VERTEX_SCALE

    glVertexPointer(3, GL_FLOAT, 0, verts);
    
    glBegin(GL_TRIANGLES);

#define STELLATE_SCALE (2.f)

    for (i=0; i<20; ++i) /* for each face of an icosahedron */
      {
        /* Place a vertex in the face */
        for (j=0; j<3; ++j)
          { /* to get the vertex in a face, average the three corners */
            /* and then scale by STELLATE_SCALE */
            verts[12+i][j] = 0;
            for (k=0; k<3; ++k)
              verts[12+i][j] += verts[ twentyfaces[i][k] ][j]*
                                    STELLATE_SCALE/3.f;
          }
        /* Render stellate */
        for (j=0; j<3; ++j) /* for each triangle in fan */
          {
            /* select a purplish color at random */
            glColor3f(FRand()*.5 + .5, FRand()*.15, FRand()*.5 + .5);
            glArrayElement(12+i);  /* fan vertex */
            glArrayElement(twentyfaces[i][j]);
            glArrayElement(twentyfaces[i][(j+1)%3]);
          }
      }

    glEnd();
  }
