/* (c) Copyright 2001 Josh Levenberg */

#include <assert.h>
#include "obstacle.h"
#include "world.h"
#include "glswmath.h"

/* Walls decide if each cell from the grid is an obstacle randomly */
/* If freq is positive, chance cell is filled is (freq-1)/freq */
/* If freq is negative, chance cell is filled is -1/freq */
/* Makes sure at least one cell is clear by randomly freeing one cell */
static unsigned Wall(int freq)
  {
    unsigned ret=0;
    unsigned shiftamt;
    assert(freq>1 || freq<-1);
    if (freq<0)
      {
        for (shiftamt=0; shiftamt < OBSTACLES_GRID; ++shiftamt)
          if (!Random(-freq))
            ret |= 1 << shiftamt;
      }
    else
      {
        for (shiftamt=0; shiftamt < OBSTACLES_GRID; ++shiftamt)
          if (Random(freq))
            ret |= 1 << shiftamt;
      }
    /* make sure we didn't get unlucky and have a full wall */
    ret &= (1<<OBSTACLES_GRID)-1 /* all bits on */
            -(1<<Random(OBSTACLES_GRID)); /* minus one bit */
    ret|=OBSTACLES_WALL;
    return ret;
  }

/* Choose two columns/rows at random */
/* First chooses one of five positions */
/* Then chooses (temp) from the remaining four */
static unsigned TwoArbitrarily()
  {
    unsigned ret=(1 << Random(5));
    unsigned temp=(1 << Random(4));
    if (temp>=ret)
      temp<<=1;
    ret |= temp;
    assert(ret!=0x1F);
    return ret;
  }

/* A standard probability distribution of row/column obstacles */
static unsigned RowsOrColumns()
  {
    unsigned ret;
    if (Random(3))
      /* Three rows or columns */
      ret=0x1F-TwoArbitrarily();
    else if (Random(4))
      /* Two rows or columns */
      ret=TwoArbitrarily();
    else
      /* Four rows or columns */
      ret = 0x1F - (1<<Random(5));
    assert(ret!=0x1F);
    if (Random(2))
      ret|=OBSTACLES_HORIZ;
    else
      ret|=OBSTACLES_VERT;
    return ret;
  }

/* Master function of this module */
/* (called from outside this module and calls all of the above functions */
/* Type of obstacle depends on level */
unsigned GenerateObstacle()
  {
    unsigned ret=0;
    switch (gLevel)
      {
        case 1:
          if (Random(2))
            /* use two wide obstacles */
            ret=3<<Random(4);
          else
            /* use one wide obstacles */
            ret=1<<Random(5);
          /* decide between horizontal and vertical Randomly */
          if (Random(2))
            ret|=OBSTACLES_HORIZ;
          else
            ret|=OBSTACLES_VERT;
          break;

        case 2:
          if (Random(3))
            /* use two wide obstacles */
            ret=3<<Random(4);
          else
            /* use three wide obstacles */
            ret=7<<Random(3);
          /* decide between horizontal and vertical Randomly */
          assert(ret!=0x1F);
          if (Random(2))
            ret|=OBSTACLES_HORIZ;
          else
            ret|=OBSTACLES_VERT;
          break;

        case 3:
          switch (Random(3))
            {
              case 0:
                /* choose a wall with about 1/4 of the blocks */
                ret=Wall(-4);
                break;
              case 1:
                /* three with a size two hole */
                ret=0xf - (0x6 << Random(2));
                break;
              case 2:
                /* two placed arbitrarily */
                ret=TwoArbitrarily();
                break;
            }
          assert(ret!=0x1F);
          if (Random(2))
            ret|=OBSTACLES_HORIZ;
          else
            ret|=OBSTACLES_VERT;
          break;
        case 4:
          if (Random(2))
            /* half the time, choose an pattern of
               either rows or columns */
            ret=RowsOrColumns();
          else
            /* choose a wall with about 1/3 of the blocks */
            ret=Wall(-3);
          break;
        default:
          if (Random(gLevel-3))
            /* wall case */
            ret=Wall(gLevel-3);
          else
            /* horizontal/veritcal case, arbitrary pattern */
            ret=RowsOrColumns();
          break;
      }
    return ret;
  }
