/* (c) Copyright 2001 Josh Levenberg */

/* In charge of anything which depends on the passage of time,
   and all changes to world state.
   Almost all gameplay rules are implemented here.
   Some is offloaded to collision detection module.
 */

#include <assert.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "physics.h"
#include "world.h"
#include "gameover.h"
#include "glutmain.h"
#include "obstacle.h"
#include "glswmath.h"
#include "text.h"
#include "collide.h"

/* Logic to determine whether obstacles are introduced */
static void SetObstacle(int which)
  {
    if (gHowFarThroughLevel + 2*NUM_TRENCH_SEGMENTS - which - 1 >=
            LEVEL_LENGTH * gLevel)
      /* Have a breather at the end of the level by not adding more 
               obstacles */
      gObstacles[which] = 0;
    else if ((gHowFarThroughLevel + which) & 1)
      /* Only add obstacles at every other trench segment */
      gObstacles[which] = GenerateObstacle();
    else
      gObstacles[which] = 0;
  }

/* Logic to determine whether gun ports are introduced */
/* Note that we interleve obstacles with gun ports */
static void SetGunPorts(int which)
  {
    if (gHowFarThroughLevel + 2*NUM_TRENCH_SEGMENTS - which - 1 >=
            LEVEL_LENGTH * gLevel)
      /* Have a breather at the end of the level by not adding more 
               obstacles */
      gGunPorts[which] = 0;
    else if ((gHowFarThroughLevel + which) & 1)
      /* Only add obstacles at every other trench segment */
      gGunPorts[which] = 0;
    else
      {
        int i;
        gGunPorts[which] = 0;
        if (Random(gLevel+1))
          for (i=0; i<gLevel; ++i)
            gGunPorts[which] |= 1 << Random(GUN_PORT_POSSIBLE);
      }
  }

/* Set up the world, so ready for game play (or demo) */
void PhysicsInit(int level)
  {
    int i;

    gPlayerPos[0] = (TRENCH_LEFT+TRENCH_RIGHT)/2;
    gPlayerPos[1] = (TRENCH_TOP+TRENCH_BOTTOM)/2;
    gPlayerPos[2] = (TRENCH_FORE+TRENCH_DEPTH)/2;

    gPlayerAim[0] = 0.5;
    gPlayerAim[1] = 0.5;

    gLevel=level;
    gScore=0;
    gHowFarThroughLevel=0;
    gSpeedUp=1; 
    /* Should be: gSpeedUp=(float) pow(gLevel, .35);
       except the only thing which calls this with level!=1
       is the demo mode, and I'm happy that goes slowly
     */

    gShootCountdown = 0;
    gCoolDown = 0;
    gFireWhenCoolDown = false;

    gShields=5;
    gShieldCountdown=0;
    
    for (i=0; i<MISSILES_MAX; ++i)
      gMissiles[i].Enabled=false;

    memset(& gObstacles[0], 0, sizeof(gObstacles[0])*(NUM_TRENCH_SEGMENTS));
    for (i=10; i<NUM_TRENCH_SEGMENTS; ++i)
      SetObstacle(i);

    InitGunPortXZArray();
    memset(& gGunPorts[0], 0, sizeof(gGunPorts[0])*(NUM_TRENCH_SEGMENTS));
    for (i=10; i<NUM_TRENCH_SEGMENTS; ++i)
      SetGunPorts(i);

    for (i=0; i<EXPLOSIONS_MAX; ++i)
      gExplosions[i].Countdown=0;

    gMessage="Starting...";
    gMessageCountdown=1.f;
  }

/* When we run into something... */
static void LoseShields()
  {
    if (gShields>0)
      {
        --gShields;
        gShieldCountdown=SHIELD_TIME;
      }
    else
      {
        /* Game over! */
        if (gDemoMode)
          /* Restart on a random level if game ends while in demo mode */
          PhysicsInit(Random(9)+1);
        else
          /* Say we are done */
          SwitchToGlutState(& gGameOverMode);
      }
  }

static void FinishLevel()
  {
    ++gLevel;
    gSpeedUp=(float) pow(gLevel, .35);
    sprintf(gMessageBuffers[0], "Level %d!", gLevel);
    gMessage=gMessageBuffers[0];
    gMessageCountdown=1.f;
    /* Bonus shield! */
    ++gShields;
  }

/* Determine when gun and how gun ports fire */
/* Fires more frequently on higher levels */
static bool MaybeFireGun(int y,int c)
  {
    /* Fire most often when at range MISSILES_OPTIMAL_RANGE */
    int offoptimal=abs(y-MISSILES_OPTIMAL_RANGE);
    int onein = 17 - gLevel;
    if (onein<4)
      onein=4;
    while (offoptimal>2)
      {
        onein*=2;
        offoptimal -= 2;
      }
    /* only fire once in onein */
    if (! Random(onein))
      {
        /* OK, we're firing the gun */
        /* Start computing the missile's parameters */
        float pos[3];
        float normal[2];
        float target[2];
        float velocity[3];
        float totaltime;

        /* compute normal (to wall) direction */
        normal[0]=0;
        normal[1]=0;
        if (c<GUN_PORT_PER_WALL)
          normal[0] = 1.f; /* left wall, normal is to right */
        else if (c<2*GUN_PORT_PER_WALL)
          normal[1] = 1.f; /* floor, normal is up */
        else
          normal[0] = -1.f; /* right wall, normal is to left */

        /* Start location for missile */
        pos[0]=gGunPortXZ[c][0];
        pos[1]=y*TRENCH_DEPTH+GUN_PORT_Y-0.8f*GUN_RADIUS;
        pos[2]=gGunPortXZ[c][1];
        /* move starting location off of wall */
        pos[0]+=normal[0]*(0.6f*GUN_RADIUS);
        pos[2]+=normal[1]*(0.6f*GUN_RADIUS);

        /* determine target */
        /* start with pos, and offset randomly in normal direction */
        target[0]=pos[0]+normal[0]*FRand()*
                  (TRENCH_RIGHT-TRENCH_LEFT-(0.6f*GUN_RADIUS));
        target[1]=pos[2]+normal[1]*FRand()*
                  (TRENCH_TOP-TRENCH_BOTTOM-(0.6f*GUN_RADIUS));
        /* then offset by +/- 1/6th of the trench in 
           the direction perpendicular to normal */
        target[0]+=normal[1]*(FRand()-.5f)*
                  (TRENCH_RIGHT-TRENCH_LEFT)/3.f;
        target[1]+=normal[0]*(FRand()-.5f)*
                  (TRENCH_TOP-TRENCH_BOTTOM)/3.f;

        /* how long we expect this missile to live */
        totaltime=(pos[1]+TRENCH_DEPTH*.5f)/
                  (PLAYER_Y_SPEED*gSpeedUp+MISSILE_Y_SPEED);
        /* velocity is distance to travel over time to get there */
        velocity[0] = (target[0]-pos[0])/totaltime;
        velocity[1] = -MISSILE_Y_SPEED;
        velocity[2] = (target[1]-pos[2])/totaltime;

        /* cap maximum deflection velocity */
        if (fabs(velocity[0])>MISSILE_XZ_SPEED*gSpeedUp)
          velocity[0] *= (float) (MISSILE_XZ_SPEED*gSpeedUp/fabs(velocity[0]));
        if (fabs(velocity[2])>MISSILE_XZ_SPEED*gSpeedUp)
          velocity[2] *= (float) (MISSILE_XZ_SPEED*gSpeedUp/fabs(velocity[2]));

        CreateMissile(pos, velocity);
        return true;
      }
    return false;
  }

/* Idea: make the cursor be twice as sensitive near the center so
   that you don't have to move the cursor all the way to the edge
   of the screen to get around an obstacle
 */
static float SteeringCorrection(float t)
  { /* Should have smooth derivative and:
          t      out
        ------------
          0       0
         .25     .1
         .5      .5
         .75     .9
          1       1
      SteeringCorrection(t) = 1 - SteeringCorrection(1-t)
      Use a cubic curve for 0 .. 1/2 and get 1/2 .. 1 by symmetry
      So for t<=1/2, SteeringCorrection(t) = a t^3 + b t^2 + c t + d
      SteeringCorrection(0) = 0     ->  d = 0
      SteeringCorrection(.25) = .1  ->  a/64 + b/16 + c/4 = 1/10
      SteeringCorrection(.5) = .5   ->  a/8  + b/4  + c/2 = 1/2
      Additional criteria: slope at .5 is 2, so
                                    ->  3a/4 + b    + c   = 2
      Solving, gives
      a = 3.2    b = -1.2   c = .8   d = 0
    */
    if (t <= .5)
      return ((3.2f*t - 1.2f)*t + .8f)*t;
    else
      return 1.f - SteeringCorrection(1.f - t);
  }

void PhysicsAdvance(double timestep)
  {
    int which_exp, which_missile;
    /* need old_y for collision detection */
    float old_y = gPlayerPos[1];
    /* --- determine steering --- */
    /* what x,z do we want to be at? */
    float xpos_targ=(float) (SteeringCorrection(gPlayerAim[0])*
                                (TRENCH_RIGHT-TRENCH_LEFT-PLAYER_WIDTH)+
                            TRENCH_LEFT+PLAYER_WIDTH*.5);
    float zpos_targ=(float) (SteeringCorrection(gPlayerAim[1])*
                                (TRENCH_TOP-TRENCH_BOTTOM-PLAYER_HEIGHT)+
                            TRENCH_BOTTOM+PLAYER_HEIGHT*.5);
    /* how far can we get? */
    float maxdist = (float) (PLAYER_XZ_SPEED*gSpeedUp*timestep);
    /* how far do we want to go? */
    float dx = xpos_targ-gPlayerPos[0];
    float dz = zpos_targ-gPlayerPos[2];
    float distsquared = dx*dx+dz*dz;
    if (distsquared <= maxdist*maxdist)
      { /* not too far, get there */
        gPlayerPos[0]=xpos_targ;
        gPlayerPos[2]=zpos_targ;
      }
    else
      { /* too far, get partway there */
        float scalefactor = (float)(maxdist/sqrt(distsquared));
        gPlayerPos[0]+=dx*scalefactor;
        gPlayerPos[2]+=dz*scalefactor;
      }

    /* move forward at constant speed */
    gPlayerPos[1] = (float) (gPlayerPos[1]+PLAYER_Y_SPEED*gSpeedUp*timestep);

    if (DetectCollision(old_y) && gShieldCountdown<=0)
      { /* Ran into an obstacle! */
        LoseShields();
      }

    /* Advance countdown for explosions */
    for (which_exp=0; which_exp<EXPLOSIONS_MAX; ++which_exp)
      if (gExplosions[which_exp].Countdown>0)
        {
          gExplosions[which_exp].Countdown -= (float) timestep;
          if (gExplosions[which_exp].Countdown<=0)
            gExplosions[which_exp].Countdown=0;
          else
            {
              int p;
              for (p=0; p<EXPLOSIONS_PARTICLES; ++p)
                gExplosions[which_exp].Particles[p].Color +=
                    (float) (FRand()*timestep/EXPLOSIONS_DURATION);
            }
        }

    /* Advance position/rotation for missiles */
    for (which_missile=0; which_missile<MISSILES_MAX; ++which_missile)
      if (gMissiles[which_missile].Enabled)
        {
          int i;
          for (i=0; i<3; ++i)
            {
              gMissiles[which_missile].Pos[i]+=(float) 
                  (timestep*gMissiles[which_missile].Velocity[i]);
            }
          gMissiles[which_missile].Rotation += (float) 
                (timestep*MISSILE_ROTATION_SPEED);
          /* disable missile if it is behind player */
          if (gMissiles[which_missile].Pos[1] < - MISSILE_RADIUS)
             gMissiles[which_missile].Enabled=false;
        }


    if (gPlayerPos[1] > TRENCH_DEPTH)
      {
        int missilecount;
        gPlayerPos[1] -= TRENCH_DEPTH;

        /* --- advance the world by one segment --- */
        /* obstacles */
        memmove(& gObstacles[0], & gObstacles[1], 
                sizeof(gObstacles[0])*(NUM_TRENCH_SEGMENTS-1));

        SetObstacle(NUM_TRENCH_SEGMENTS-1);

        /* gun ports */
        memmove(& gGunPorts[0], & gGunPorts[1], 
                sizeof(gGunPorts[0])*(NUM_TRENCH_SEGMENTS-1));

        SetGunPorts(NUM_TRENCH_SEGMENTS-1);

        /* explosions */
        for (which_exp=0; which_exp<EXPLOSIONS_MAX; ++which_exp)
          if (gExplosions[which_exp].Countdown > 0)
            {
              gExplosions[which_exp].Pos[1] -= TRENCH_DEPTH;
              /* check to see if we passed the explosion */
              if (gExplosions[which_exp].Pos[1]<0)
                gExplosions[which_exp].Countdown=0;
            }

        /* missiles */
        missilecount=0;
        for (which_missile=0; which_missile<MISSILES_MAX; ++which_missile)
          if (gMissiles[which_missile].Enabled)
            {
              gMissiles[which_missile].Pos[1] -= TRENCH_DEPTH;
              if (gMissiles[which_missile].Pos[1] < - MISSILE_RADIUS)
                gMissiles[which_missile].Enabled=false;
              else
                ++missilecount;
            }
        /* decide if any gun ports will fire a missile */
        if (missilecount<MISSILES_MAX && gLevel>1)
          {
            int y;
            /* scan through missile ports not at point blank range */
            for (y=2; y<NUM_TRENCH_SEGMENTS; ++y)
              if (gGunPorts[y])
                {
                  int c;
                  for (c=0; c<GUN_PORT_POSSIBLE; ++c)
                    if (gGunPorts[y] & (1 << c))
                      /* found a gun port, now determine if it fires */
                      if (MaybeFireGun(y,c))
                        { /* if fired, abort loop; so we only
                             fire one at a time */
                          ++missilecount;
                          y=NUM_TRENCH_SEGMENTS;
                          break;
                        }
                }
          }

        ++gHowFarThroughLevel;
        /* --- rewards for making progress --- */
        gScore += gLevel;
        if (gHowFarThroughLevel >= LEVEL_LENGTH * gLevel)
          {
            /* Cleared a level! */
            FinishLevel();
            gHowFarThroughLevel=0;
          }
      }

    /* Shot count down */
    if (gShootCountdown>0)
      {
        /* check to see if we hit something */
        ComputeShotDestination();

        gShootCountdown -= (float) timestep;
        /* firing done yet? */
        if (gShootCountdown<=0)
          { /* yes, cool down */
            gShootCountdown=0;
            gCoolDown=COOL_DOWN_TIME;
          }
      }
    else if (gCoolDown>0)
      {
        gCoolDown -= (float) timestep;
        if (gCoolDown<=0)
          {
            gCoolDown=0;
            if (gFireWhenCoolDown)
              {
                gFireWhenCoolDown=false;
                PhysicsFire();
              }
          }
      }

    /* If just used shield, invulnerable for gShieldCountdown */
    if (gShieldCountdown>0)
      {
        gShieldCountdown -= (float) timestep;
        if (gShieldCountdown<=0)
          gShieldCountdown=0;
      }

    /* See if we should clear the message */
    if (gMessageCountdown>0)
      {
        gMessageCountdown -= (float) timestep/MESSAGE_TIME;
        if (gMessageCountdown<=0)
          {
            gMessageCountdown=0;
            gMessage=0;
          }
      }
  }

void PhysicsFire()
  {
    if (gShootCountdown>0)
      ; /* already firing, ignore shot */
    else if (gCoolDown)
      /* fire as soon as we can */
      gFireWhenCoolDown = true;
    else
      { /* Start firing! */
        gShootCountdown = SHOOT_TIME;
        gShootAim[0]=gPlayerAim[0];
        gShootAim[1]=gPlayerAim[1];
        ComputeShotDestination();
      }
  }

/* in response to mouse movement */
void PhysicsAim(float x, float z)
  {
    assert(x>=0);
    assert(x<=1);
    assert(z>=0);
    assert(z<=1);
    gPlayerAim[0]=x;
    gPlayerAim[1]=z;
  }
