/* (c) Copyright 2001 Josh Levenberg */

/* Contains all world definitions and state variables */

#ifndef __GL_SW_WORLD_H
#define __GL_SW_WORLD_H

/* for "bool" */
#include "types.h"

/* Coordinate system:
   +x right
   +y forward
   +z up
 */

/* Are we in the intro screen? -------------------------------------- */

extern bool gDemoMode;

/* Player stats ----------------------------------------------------- */

/* gPlayerPos: where in worldspace the player is. 
   Note that whenever it exceeds TRENCH_DEPTH, a trench segment
   is removed and the world is shifted by TRENCH_DEPTH */
extern float gPlayerPos[3];

/* gPlayerAim: where the mouse is pointing */
extern float gPlayerAim[2];

/* Collision rectangle.  Also used to position "quad lasers". */
#define PLAYER_HEIGHT   (.1f)
#define PLAYER_WIDTH    (.125f)

/* SPEEDs are in units per second */
#define PLAYER_Y_SPEED  (1.0f)
#define PLAYER_XZ_SPEED (.5f)

/* Where the player's collision box is along the Y axis 
   Morally, this should be 0, 
   but a small positive value is more intuitive.
 */
#define PLAYER_COLLISION_Y (0.08f)

extern int gLevel;
extern unsigned gScore;

/* How many segments of trench have we passed so far
   this level */
extern int gHowFarThroughLevel;

/* Player forward speed and steering speed
   are multiplied by the current value of gSpeedUp */
extern float gSpeedUp;

/* Player shooting stats -------------------------------------------- */
/* Shots fire with the same aim (stored in gShootAim)
   over a period of "SHOOT_TIME".  After firing, weapons can't be fired
   until after COOL_DOWN_TIME has elapsed (though clicking during this
   time will cause the player to fire as soon as the cool down is 
   finished */
/* (Times are in seconds) */
#define SHOOT_TIME      (.2f)
#define COOL_DOWN_TIME  (.1f)
/* how close are we to finishing shooting/cooling down? */
extern float gShootCountdown;
extern float gCoolDown;
/* did the player fire during the cool down period? */
extern bool gFireWhenCoolDown;
/* where was the last shot aimed? */
extern float gShootAim[2];
/* where the quad lasers are converging */
extern float gShotDest[3];

/* Player shield stats ---------------------------------------------- */
/* Shields are like extra lives -- they protect the player from 
   collisions.  Note that the game does not stop when you hit 
   something -- you have SHIELD_TIME of grace period where you
   can't be damaged.
 */
extern int gShields;
/* If we are invulnerable this is >0 and indicates how 
   long we have left */
extern float gShieldCountdown;

#define SHIELD_TIME     (1.6f)

/* Trench statistics ------------------------------------------------ */
enum { NUM_TRENCH_SEGMENTS=20 };  /* visible at once */
enum { LEVEL_LENGTH = 40 };       /* times gLevel */
#define TRENCH_TOP     (1.0f)
#define TRENCH_BOTTOM  (0.0f)
#define TRENCH_LEFT    (0.0f)
#define TRENCH_RIGHT   (1.25f)
#define TRENCH_FORE    (0.0f)
#define TRENCH_DEPTH   (1.0f)

/* Obstacle state --------------------------------------------------- */
/* Colliding with obstacles (green in the game) hurts the player.
   Everything else passes through them.
   They are horizontal, vertical, or a wall (grid).
   Obstacles are divided into a grid OBSTACLES_ROWS by OBSTACLES_COLUMNS
   wide.
 */
enum { OBSTACLES_ROWS=5 };
enum { OBSTACLES_COLUMNS=5 };
enum { OBSTACLES_GRID=OBSTACLES_ROWS*OBSTACLES_COLUMNS };
/* since we are on a 32 bit machine and the high two bits are reserved
   for the obstacle type, OBSTACLES_GRID should never exceed 30 */

enum { OBSTACLES_NONE  = 0x00000000,
       OBSTACLES_HORIZ = 0x40000000,
       OBSTACLES_VERT  = 0x80000000,
       OBSTACLES_WALL  = 0xC0000000,
       OBSTACLES_TYPE_MASK = 0xC0000000 };

extern unsigned long gObstacles[NUM_TRENCH_SEGMENTS];
/* if 0, no obstacle in this trench segment */
/* if &OBSTACLES_TYPE_MASK==OBSTACLES_HORIZ then there are 
      horizontal obstacles, the low OBSTACLES_ROWS bits indicate 
      which are present
 */
/* if &OBSTACLES_TYPE_MASK==OBSTACLES_VERT then there are 
      vertical obstacles, the low OBSTACLES_COLUMNS bits indicate 
      which are present
 */
/* if &OBSTACLES_TYPE_MASK==OBSTACLES_WALL then there is a
      grid of obstacles, the low OBSTACLES_ROWS*OBSTACLES_COLUMNS bits 
      indicate which are present
 */

/* OBSTACLE_Y gives the Y position within a trench segment of 
   obstacles */
#define OBSTACLE_Y    (0.5f)

/* the dimensions of each grid cell */
#define OBSTACLE_WIDE ((TRENCH_RIGHT-TRENCH_LEFT)/OBSTACLES_COLUMNS)
#define OBSTACLE_HIGH ((TRENCH_TOP-TRENCH_BOTTOM)/OBSTACLES_ROWS)

/* Guns state ------------------------------------------------------- */
/* Gun ports, or "guns," are on the trench walls or floor.  They will
   occationally shoot missiles (see below).  They can be shot.
 */
enum { GUN_PORT_PER_WALL=3 };
enum { GUN_PORT_POSSIBLE=3*GUN_PORT_PER_WALL };
/* since we're using 16 bit bitmasks below, GUN_PORT_POSSIBLE
   should be at most 16 */
enum { GUN_PORT_MASK=((1 << GUN_PORT_POSSIBLE)-1) };
/* Score for shooting one */
enum { GUN_PORT_SCORE=15 };
/* Dimensions */
#define GUN_RADIUS (0.15f)
/* Y position within a trench segment */
#define GUN_PORT_Y (0.5f)

/* a bitmask determines which of the GUN_PORT_POSSIBLE gun ports
   are present in any given trench segement
 */
extern unsigned short gGunPorts[NUM_TRENCH_SEGMENTS];
/* XZ positions of gun ports within trench segments */
extern float gGunPortXZ[GUN_PORT_POSSIBLE][2];

/* sets gGunPortXZ */
void InitGunPortXZArray(void);

/* Missile stats ---------------------------------------------------- */
/* Missiles reduce your shields if they hit you.  They may be shot down.
   Once fired (by a gun port), they move with constant velocity and
   rotate with constant angular velocity.
 */
/* max number of missiles present at once */
enum { MISSILES_MAX=6 };
/* when missiles are most likely to fire */
enum { MISSILES_OPTIMAL_RANGE=9 };
/* Score for shooting one */
enum { MISSILE_SCORE=10 };
/* Radius for collision purposes */
#define MISSILE_RADIUS (0.15f)
/* Rotational speed in degrees per second */
#define MISSILE_ROTATION_SPEED (60.f)
/* Speed in world units per second */
#define MISSILE_Y_SPEED (.5f)
#define MISSILE_XZ_SPEED (.2f)
/* Note: MISSILE_XZ_SPEED is multiplied by the 
   current value of gSpeedUp */

typedef struct 
  {
    float Pos[3];       /* World position */
    float Velocity[3];  /* World units per second */
    float Axis[3];      /* Axis of rotation */
    float Rotation;     /* Degrees rotated around above axis */
    bool  Enabled;      /* true if this missile slot is active */
  } MissileInfo;
extern MissileInfo gMissiles[MISSILES_MAX];

/* Create missile will silently do nothing if all the
   missile slots are in use.  Otherwise it will create a new missile
   in an unused slot.  It randomly sets the rotation axis and
   the initial rotation angle.
 */
void CreateMissile(float pos[3], float vel[3]);

/* Explosion stats -------------------------------------------------- */
/* Explosions are there to look cool, but they do have a game play 
   effect: they absorb the players shots.  Explosions are used
   anytime a gun port or missile is destroyed.  A simple particle
   system is used for the effect.
 */
/* radius for collision purposes in world units */
#define EXPLOSIONS_RADIUS       (0.2f)
/* in seconds */
#define EXPLOSIONS_DURATION     (0.7f)
/* largest size of particles in world units */
#define EXPLOSION_PARTICLE_SIZE (0.03f)

/* most simultaneous explosions */
enum { EXPLOSIONS_MAX=4 };
/* number of particles in each explosion */
enum { EXPLOSIONS_PARTICLES=50 };

typedef struct
  {
    float Vect[3]; /* this vector is scaled with time and added to 
                      the explosion's center position to get the
                      particle's position */
    float Color;   /* from 0 to 1, white to yellow to red */
  } ExplosionParticle;

typedef struct 
  {
    float Pos[3];     /* center of the explosion */
    float Countdown;  /* time to live */
    ExplosionParticle Particles[EXPLOSIONS_PARTICLES];
  } ExplosionInfo;
ExplosionInfo gExplosions[EXPLOSIONS_MAX];

/* CreateExplosion replaces the explosion with the smallest
   remaining time (which is 0 unless there are 4 simultaneous explosions)
   It initializes the count down to EXPLOSIONS_DURATION and creates
   the particles.  dir is used to make the particles go away from
   a wall or the floor when a gun port is shot.  Use EXPLODE_ALL
   if you want the particles to go in all directions.
 */

typedef enum { EXPLODE_ALL, EXPLODE_RIGHT, EXPLODE_UP, EXPLODE_LEFT } 
  ExplodeDirection;
void CreateExplosion(float location[3], ExplodeDirection dir);

/* Message ---------------------------------------------------------- */
/* Message to display at the top during game play.  Used at the start
   of the game and when the level advances.  gMessage can safely be
   pointed at gMessageBuffers[0] (see text.h).
 */
extern const char *gMessage;
/* Unlike the other countdowns above, 
       ** gMessageCountdown is NOT in seconds! **
   It starts at 1 and counts down to 0 over the span of MESSAGE_TIME
   seconds.
 */
extern float gMessageCountdown;

#define MESSAGE_TIME    (6.f)

#endif /* #ifndef __GL_SW_WORLD_H */
